import numpy as np
import pandas as pd
from scipy.stats import dirichlet
import dirichlet as drm
from fmvmm.utils.utils_dmm import (kmeans_init, gmm_init, kmeans_init_adv, gmm_init_adv, random_init,
                             dmm_loglikelihood, dmm_responsibilities, dmm_pi_estimate)


from fmvmm.utils.utils_mixture import (mixture_clusters)


def estimate_alphas(data_cwise, alpha_not, method):
    alpha_new = []
    for t in range(len(data_cwise)):
        if np.array(data_cwise[t]).ndim == 1:
            alpha_new_temp = alpha_not[t]
        elif method == 'fixedpoint':
            alpha_new_temp = drm.mle(
                np.array(data_cwise[t]), method="fixedpoint", maxiter=9223372036854775807)
        else:
            alpha_new_temp = drm.mle(np.array(data_cwise[t]))

        alpha_new.append(alpha_new_temp)
    return alpha_new


class DMM:

    def __init__(self, number_of_clusters, sample, method="meanprecision", initialization="KMeans", tol=0.0001,show_loglikelihood_diff=False):
        self.number_of_clusters = number_of_clusters
        self.sample = sample
        self.method = method
        self.initialization = initialization
        self.tol = tol
        self.show_loglikelihood_diff=show_loglikelihood_diff
        self.k = self.number_of_clusters
        data = self.sample
        self.p = len(data.columns)
        self.n = len(data)

        self.data_lol = data.values.tolist()
        if self.initialization == "KMeans":
            self.pi_not, self.alpha_not = kmeans_init(data, self.k)
        elif self.initialization == "GMM":
            self.pi_not, self.alpha_not = gmm_init(data, self.k)
        elif self.initialization == "random":
            self.pi_not, self.alpha_not = random_init(data, self.k)

        self.alpha_temp = self.alpha_not
        self.pi_temp = self.pi_not

    def fit(self):
        log_like_diff = 5
        while log_like_diff > self.tol:

            log_likelihood_old = dmm_loglikelihood(
                self.pi_temp, self.alpha_temp, self.data_lol)
            gamma_temp_ar, gamma_matrix = dmm_responsibilities(
                self.pi_temp, self.alpha_temp, self.data_lol)
            pi_new = dmm_pi_estimate(gamma_temp_ar)

            cluster, data_cwise = mixture_clusters(gamma_matrix, self.data_lol)

            alpha_new = estimate_alphas(
                data_cwise, self.alpha_not, self.method)

            log_likelihood_new = dmm_loglikelihood(
                pi_new, alpha_new, self.data_lol)

            log_like_diff = abs(log_likelihood_new-log_likelihood_old)/log_likelihood_old
            estimated_mean = []
            for a in alpha_new:
                mean_temp = [b/np.sum(a) for b in a]
                estimated_mean.append(mean_temp)

            self.alpha_temp = alpha_new
            self.pi_temp = pi_new
            if self.show_loglikelihood_diff:
                print(log_like_diff)
        self.pi_new = pi_new
        self.alpha_new = alpha_new
        self.estimated_mean = estimated_mean
        self.cluster = cluster
        self.data_cwise = data_cwise
        self.gamma_temp_ar = gamma_temp_ar
        self.gamma_matrix = gamma_matrix
        self.log_likelihood_new = log_likelihood_new
        print("Hard DMM Fitting Done Successfully")

    def get_params(self):
        # print("The estimated pi values are ", self.pi_new)
        # print("The estimated alpha values are ", [l.tolist() for l in self.alpha_new])

        return self.pi_new, self.alpha_new

    def predict(self):
        return self.cluster

    def predict_new(self, x):
        data_lol = x.values.tolist()
        cluster, _ = mixture_clusters(self.gamma_matrix, data_lol)

        return cluster

    def estimated_mean(self):
        estimated_mean = pd.DataFrame(self.estimated_mean)
        return estimated_mean

    def responsibilities(self):

        return self.gamma_temp_ar

    def clustered_data(self):
        return pd.DataFrame(self.data_cwise).transpose()

    def bic(self):

        return ((self.k-1)+(self.k*self.p))*np.log(self.n) - 2*(self.log_likelihood_new)

    def aic(self):

        return 2*((self.k-1)+(self.k*self.p)) - 2*(self.log_likelihood_new)

    def icl(self):
        entropy_s=[]
        for i in self.gamma_matrix:
            for j in i:
                    ent_temp=j*np.log1p(j)
                    entropy_s.append(ent_temp)



        entropy=np.sum(entropy_s)

        return ((self.k-1)+(self.k*self.p))*np.log(self.n) - 2*(self.log_likelihood_new) - entropy



import numpy as np
from sklearn.metrics import confusion_matrix
import copy
import matplotlib.pyplot as plt


def acc_check(tru, obs):
    lom = confusion_matrix(tru, obs).tolist()

    match = [max(lom[0])]
    max_prob_c = [lom[0].index(max(lom[0]))]
    for i in range(1, len(lom)):
        for j in max_prob_c:
            del lom[i][j]
        max_prob_c_temp = lom[i].index(max(lom[i]))
        max_prob_c.append(max_prob_c_temp)
        temp_match = max(lom[i])
        match.append(temp_match)

    match = np.sum([max(i) for i in lom])
    acc = match/len(tru)

    return acc


def prec_check(tru, obs):
    lom = confusion_matrix(tru, obs).tolist()
    lom_c = copy.deepcopy(lom)

    match = [max(lom_c[0])]
    max_prob_c = [lom[0].index(max(lom_c[0]))]
    for i in range(1, len(lom_c)):
        for j in sorted(max_prob_c, reverse=True):
            del lom_c[i][j]
        max_prob_c_temp = lom[i].index(max(lom_c[i]))
        max_prob_c.append(max_prob_c_temp)
        temp_match = max(lom_c[i])
        match.append(temp_match)
    lom = confusion_matrix(tru, obs).tolist()
    prec = []
    for k in range(len(match)):
        temp_prec = match[k]/np.sum(lom[k])
        prec.append(temp_prec)

    return np.mean(prec)


def rec_check(tru, obs):
    lom = confusion_matrix(tru, obs).tolist()
    lom_c = copy.deepcopy(lom)

    match = [max(lom_c[0])]
    max_prob_c = [lom_c[0].index(max(lom_c[0]))]
    for i in range(1, len(lom_c)):
        for j in sorted(max_prob_c, reverse=True):
            del lom_c[i][j]
        max_prob_c_temp = lom[i].index(max(lom_c[i]))
        max_prob_c.append(max_prob_c_temp)
        temp_match = max(lom_c[i])
        match.append(temp_match)

    lom = confusion_matrix(tru, obs).tolist()
    rec = []
    for k in range(len(match)):
        temp_rec = match[k]/np.sum([lom[l][max_prob_c[k]]
                                   for l in range(len(lom))])
        rec.append(temp_rec)

    return np.mean(rec)


def f_score(tru, obs):
    prec = prec_check(tru, obs)
    rec = rec_check(tru, obs)
    f_sc = 2/(prec**(-1)+rec**(-1))

    return f_sc


def mixture_clusters(gamma_matrix, data_lol):
    n = len(data_lol)
    k = len(gamma_matrix[0])
    cluster = []
    data_cwise = [[] for i in range(k)]
    for h in range(n):
        max_prob_c = gamma_matrix[h].index(max(gamma_matrix[h]))
        data_cwise[max_prob_c].append(data_lol[h])
        cluster.append(max_prob_c)

    return cluster, data_cwise

def plot_correlation_heatmap(covariance_matrix, variable_names):
    # Calculate correlation matrix from the covariance matrix
    correlation_matrix = np.corrcoef(covariance_matrix)

    # Create the heatmap plot
    fig, ax = plt.subplots()
    im = ax.imshow(correlation_matrix, cmap='RdYlBu', vmin=-1, vmax=1)

    # Set x-axis and y-axis tick labels
    ax.set_xticks(np.arange(len(variable_names)))
    ax.set_yticks(np.arange(len(variable_names)))
    ax.set_xticklabels(variable_names)
    ax.set_yticklabels(variable_names)

    # Rotate x-axis tick labels for better readability
    plt.setp(ax.get_xticklabels(), rotation=45, ha='right', rotation_mode='anchor')

    # Add colorbar
    cbar = ax.figure.colorbar(im, ax=ax)
    cbar.set_label('Correlation')

    # Set plot title
    ax.set_title('Correlation Heatmap')

    # Show the plot
    plt.show()



import sys

import numpy as np
from numpy import (
    asanyarray,
    exp
)
from numpy.linalg import norm
from scipy.special import polygamma, psi
from scipy.special import digamma
import pandas as pd
import copy
from fmvmm.utils.utils_dmm import (kmeans_init, gmm_init, kmeans_init_adv, gmm_init_adv, random_init,
                             dmm_loglikelihood, dmm_responsibilities, dmm_pi_estimate)

from fmvmm.utils.utils_mixture import (mixture_clusters)
MAXINT = sys.maxsize
euler = -1 * psi(1)  # Euler-Mascheroni constant


class NotConvergingError(Exception):
    """Error when a successive approximation method doesn't converge
    """
    pass


def _ipsi(y, tol=1.48e-9, maxiter=10):
    """Inverse of psi (digamma) using Newton's method. For the purposes
    of Dirichlet MLE, since the parameters a[i] must always
    satisfy a > 0, we define ipsi :: R -> (0,inf).

    Parameters
    ----------
    y : (K,) shape array
        y-values of psi(x)
    tol : float
        If Euclidean distance between successive parameter arrays is less than
        ``tol``, calculation is taken to have converged.
    maxiter : int
        Maximum number of iterations to take calculations. Default is 10.

    Returns
    -------
    (K,) shape array
        Approximate x for psi(x)."""
    y = asanyarray(y, dtype="float")
    x0 = np.piecewise(
        y,
        [y >= -2.22, y < -2.22],
        [(lambda x: exp(x) + 0.5), (lambda x: -1 / (x + euler))],
    )
    for i in range(maxiter):
        x1 = x0 - (psi(x0) - y) / _trigamma(x0)
        if norm(x1 - x0) < tol:
            return x1
        x0 = x1
    raise NotConvergingError(
        f"Failed to converge after {maxiter} iterations, " f"value is {x1}")


def _trigamma(x):
    return polygamma(1, x)


def dirichlet_mix_mle(x: np.ndarray, gamma: np.ndarray, alpha_init):

    p = x.shape[1]
    n = x.shape[0]
    k = gamma.shape[1]
    x_lol = x.tolist()
    alpha_not = alpha_init
    N_js = [np.sum(gamma[:, l]) for l in range(k)]
    alpha_old = alpha_not
    alpha_all = []
    for o in range(k):

        log_x = []
        for f in range(p):
            log_x_temp = (
                1/N_js[o])*np.sum([np.log(x_lol[j][f])*gamma[j, o] for j in range(n)])
            log_x.append(log_x_temp)
        alpha_new = []
        for i in range(p):
            diff = 5
            t = copy.deepcopy(alpha_old[o][i])
            t1 = copy.deepcopy(alpha_old[o][i])
            di_alpha_not = digamma(t)
            di_alpha_old = di_alpha_not
            beta = np.sum(alpha_old[o])-t
            if beta > 500:
                alpha_new_temp = _ipsi(
                    np.log(beta)+log_x[i], tol=1.48e-9, maxiter=10000)
            else:

                while diff > 0.0001:
                    di_alpha_sum = digamma(np.sum(alpha_old[o])-t + t1)
                    di_alpha_new = di_alpha_sum + log_x[i]
                    t1 = _ipsi(di_alpha_new, tol=1.48e-9, maxiter=1000)
                    diff = abs(di_alpha_new - di_alpha_old)
                    di_alpha_old = di_alpha_new
                    alpha_new_temp = t1
            alpha_new.append(alpha_new_temp)
        alpha_all.append(alpha_new)

    return alpha_all


def dirichlet_mean_precision_mle(x: np.ndarray, gamma: np.ndarray, alpha_init):
    alpha_old = alpha_init
    diff = 5
    norm0 = 5
    while diff > 0.0001:
        alpha_1 = _fit_s(x, gamma, alpha_old)
        alpha_2 = _fit_m(x, gamma, alpha_1)
        norm1 = norm(np.array(alpha_2)-np.array(alpha_old))
        diff = abs(norm1-norm0)
        #print("alpha_diff",diff)
        norm0 = norm1
        alpha_old = alpha_2
    return alpha_2


def _fit_s(x: np.ndarray, gamma: np.ndarray, alpha_old):

    p = x.shape[1]
    n = x.shape[0]
    k = gamma.shape[1]
    x_lol = x.tolist()
    N_js = [np.sum(gamma[:, l]) for l in range(k)]
    sj_old = [np.sum(alpha_old[i]) for i in range(len(alpha_old))]
    mj_old = [(np.array(alpha_old[i])/sj_old[i]).tolist()
              for i in range(len(alpha_old))]
    alpha_new_all = []
    for o in range(k):
        diff = 5
        sjold = sj_old[o]
        while diff > 0.0001:
            cs = []
            for i in range(n):
                for j in range(p):
                    c_temp = mj_old[o][j]*gamma[i, o]*np.log(x_lol[i][j])
                    cs.append(c_temp)
            c = np.sum(cs)
            bs = []
            for i in range(p):
                b_temp = mj_old[o][i]*digamma(sjold*mj_old[o][i])
                bs.append(b_temp)
            b = N_js[o]*np.sum(bs)
            a = N_js[o]*digamma(sjold)
            firststdif = a-b+c
            d = N_js[o]*_trigamma(sjold)
            es = []
            for i in range(p):
                es_temp = (mj_old[o][i]**2)*_trigamma(sjold*mj_old[o][i])
                es.append(es_temp)
            e = N_js[o]*np.sum(es)
            seconddif = d-e
            if seconddif < 0:
                sj_new = sjold - firststdif/seconddif
            elif firststdif + sjold*seconddif < 0:
                sj_new = 1/((1/sjold)+(1/sjold**2)*(1/seconddif)*firststdif)
            else:
                raise NotConvergingError(f"Unable to update s from {sjold}")
            diff = abs(sj_new-sjold)
            #print("s_diff",diff)
            sjold = sj_new
        alphaj_new = [sj_new*mj_old[o][i] for i in range(p)]
        alpha_new_all.append(alphaj_new)
    return alpha_new_all


def _fit_m(x: np.ndarray, gamma: np.ndarray, alpha_old):
    p = x.shape[1]
    n = x.shape[0]
    k = gamma.shape[1]
    x_lol = x.tolist()
    N_js = [np.sum(gamma[:, l]) for l in range(k)]
    sj_old = [np.sum(alpha_old[i]) for i in range(len(alpha_old))]
    mj_old = [(np.array(alpha_old[i])/sj_old[i]).tolist()
              for i in range(len(alpha_old))]
    alpha_new_all = []

    for o in range(k):

        sjold = sj_old[o]
        mjold = mj_old[o]
        alpha_j_old = alpha_old[o]
        m6 = []
        for s in range(p):
            m5 = np.sum([gamma[i, o]*np.log(x_lol[i][s])
                        for i in range(n)])/N_js[o]
            m6.append(m5)
        diff = 5
        norm0 = 5
        while diff > 0.0001:
            alpha_j_new = []
            for j in range(p):
                m1 = m6[j]
                m2 = [mjold[q]*m6[q]for q in range(p)]
                m3 = np.sum(m2)
                m4 = np.sum([mjold[r]*digamma(sjold*mjold[r])
                            for r in range(p)])
                alpha_jm_new = _ipsi(m1-m3+m4, tol=1.48e-9, maxiter=1000)
                alpha_j_new.append(alpha_jm_new)
            alpha_j_new = np.array(alpha_j_new) / \
                np.array(alpha_j_new).sum() * sjold
            norm1 = norm(alpha_j_new-alpha_j_old)
            diff = abs(norm1-norm0)
            #print("m_diff",diff)
            norm0 = norm1
            mjold = alpha_j_new/alpha_j_new.sum()
        alpha_new_all.append(alpha_j_new.tolist())
    return alpha_new_all


def estimate_alphas(data: np.ndarray, gamma_temp_ar, alpha_temp, method):
    if method == "fixediteration":
        alpha_new = dirichlet_mix_mle(data, gamma_temp_ar, alpha_temp)
    else:
        alpha_new = dirichlet_mean_precision_mle(
            data, gamma_temp_ar, alpha_temp)
    return alpha_new


class DMM_Soft:

    def __init__(self, number_of_clusters, sample, method="meanprecision", initialization="KMeans", tol=0.0001,show_loglikelihood_diff=False):
        self.number_of_clusters = number_of_clusters
        self.sample = sample
        self.method = method
        self.initialization = initialization
        self.tol = tol
        self.show_loglikelihood_diff=show_loglikelihood_diff
        self.k = self.number_of_clusters
        data = self.sample
        self.p = len(data.columns)
        self.n = len(data)

        self.data_lol = data.values.tolist()
        np.random.seed(0)
        if self.initialization == "KMeans":
            self.pi_not, self.alpha_not = kmeans_init(data, self.k)
        elif self.initialization == "GMM":
            self.pi_not, self.alpha_not = gmm_init(data, self.k)
        elif self.initialization == "KMeans Adv":
            self.pi_not, self.alpha_not = kmeans_init_adv(data, self.k)
        elif self.initialization == "GMM Adv":
            self.pi_not, self.alpha_not = gmm_init_adv(data, self.k)
        elif self.initialization == "random":
            self.pi_not, self.alpha_not = random_init(data, self.k)

        self.alpha_temp = self.alpha_not
        self.pi_temp = self.pi_not

    def fit(self):
        log_like_diff = 5
        while log_like_diff > self.tol:
            log_likelihood_old = dmm_loglikelihood(
                self.pi_temp, self.alpha_temp, self.data_lol)
            gamma_temp_ar, gamma_matrix = dmm_responsibilities(
                self.pi_temp, self.alpha_temp, self.data_lol)
            pi_new = dmm_pi_estimate(gamma_temp_ar)

            cluster, data_cwise = mixture_clusters(gamma_matrix, self.data_lol)

            alpha_new = estimate_alphas(
                self.sample.to_numpy(), gamma_temp_ar, self.alpha_temp, self.method)

            log_likelihood_new = dmm_loglikelihood(
                pi_new, alpha_new, self.data_lol)

            log_like_diff = abs(log_likelihood_new-log_likelihood_old)/log_likelihood_old
            estimated_mean = []
            for a in alpha_new:
                mean_temp = [b/np.sum(a) for b in a]
                estimated_mean.append(mean_temp)

            self.alpha_temp = alpha_new
            self.pi_temp = pi_new
            if self.show_loglikelihood_diff:
                print(log_like_diff)
        self.pi_new = pi_new
        self.alpha_new = alpha_new
        self.estimated_mean = estimated_mean
        self.cluster = cluster
        self.data_cwise = data_cwise
        self.gamma_temp_ar = gamma_temp_ar
        self.gamma_matrix = gamma_matrix
        self.log_likelihood_new = log_likelihood_new
        print("Soft DMM Fitting Done Successfully")

    def get_params(self):
        #print("The estimated pi values are ", self.pi_new)
        #print("The estimated alpha values are ", self.alpha_new)

        return self.pi_new, self.alpha_new

    def predict(self):
        return self.cluster

    def predict_new(self, x):
        data_lol = x.values.tolist()
        cluster, _ = mixture_clusters(self.gamma_matrix, data_lol)

        return cluster

    def estimated_mean(self):
        estimated_mean = pd.DataFrame(self.estimated_mean)
        return estimated_mean

    def responsibilities(self):

        return self.gamma_temp_ar

    def clustered_data(self):
        return pd.DataFrame(self.data_cwise).transpose()

    def bic(self):

        return ((self.k-1)+(self.k*self.p))*np.log(self.n) - 2*(self.log_likelihood_new)

    def aic(self):

        return 2*((self.k-1)+(self.k*self.p)) - 2*(self.log_likelihood_new)

    def icl(self):
        entropy_s=[]
        for i in self.gamma_matrix:
            for j in i:
                    ent_temp=j*np.log1p(j)
                    entropy_s.append(ent_temp)



        entropy=np.sum(entropy_s)

        return ((self.k-1)+(self.k*self.p))*np.log(self.n) - 2*(self.log_likelihood_new) - entropy



from setuptools import setup, find_packages

project_url = {
  'Link 1': 'https://github.com/samyajoypal/fmvmm'
}

setup(
    name='fmvmm',
    version='1.0.12',
    author='Samyajoy Pal',
    author_email='palsamyajoy@gmail.com',
    description="flexible multivariate mixture model",
    long_description="mixture model with different distributions",
    project_urls = project_url,
    packages=find_packages(),
    install_requires=[
        'numpy',
        'scipy',
        'pandas',
        'dirichlet',
        'sklearn',
        'conorm'
    ],
    entry_points={
        'console_scripts': [
            'my_command = fmvmm.command:main',
        ],
    },
)



import numpy as np
from sklearn.cluster import KMeans
from sklearn import mixture
from scipy.stats import dirichlet
import math
import pandas as pd
import conorm


def closure(d_mat):
    d_mat = np.atleast_2d(d_mat)
    if np.any(d_mat < 0):
        raise ValueError("Cannot have negative proportions")
    if d_mat.ndim > 2:
        raise ValueError("Input matrix can only have two dimensions or less")
    if np.all(d_mat == 0, axis=1).sum() > 0:
        raise ValueError("Input matrix cannot have rows with all zeros")
    d_mat = d_mat / d_mat.sum(axis=1, keepdims=True)
    return d_mat.squeeze()

def multiplicative_replacement(d_mat, delta=None):
    d_mat = closure(d_mat)
    z_mat = (d_mat == 0)

    num_feats = d_mat.shape[-1]
    tot = z_mat.sum(axis=-1, keepdims=True)

    if delta is None:
        delta = (1. / num_feats)**2

    zcnts = 1 - tot * delta
    if np.any(zcnts) < 0:
        raise ValueError('The multiplicative replacement created negative '
                         'proportions. Consider using a smaller `delta`.')
    d_mat = np.where(z_mat, delta, zcnts * d_mat)
    return d_mat.squeeze()


def kmeans_init(data, k):
    n = len(data)
    p = len(data.columns)
    kmeans = KMeans(n_clusters=k, random_state=1)
    kmeans.fit(data)
    mu_not = kmeans.cluster_centers_
    #alsum=60
    alsum = p*5
    alpha_not = mu_not*alsum
    data_lol = data.values.tolist()
    data_cwise = [[] for i in range(k)]
    for i in range(n):
        data_cwise[kmeans.labels_[i]].append(data_lol[i])
    pi_not = [len(data_cwise[m])/n for m in range(k)]
    return pi_not, alpha_not


def gmm_init(data, k):
    n = len(data)
    clf = mixture.GaussianMixture(n_components=k, covariance_type='full')
    clf.fit(data)
    mu_not = clf.means_
    alsum = 60
    alpha_not = mu_not*alsum
    data_lol = data.values.tolist()
    data_cwise = [[] for i in range(k)]
    for i in range(n):
        data_cwise[clf.predict(data)[i]].append(data_lol[i])
    pi_not = [len(data_cwise[m])/n for m in range(k)]
    return pi_not, alpha_not


def random_init(data, k, random_seed=0):
    np.random.seed(random_seed)
    p = len(data.columns)
    alpha_not = []
    for h in range(k):
        alpha_not_temp = np.random.uniform(0, 50, p)
        alpha_not.append(alpha_not_temp)
    pi_not = sum(np.random.dirichlet([0.5 for i in range(k)], 1).tolist(), [])

    return pi_not, alpha_not


def kmeans_init_adv(data, k):
    n = len(data)
    kmeans = KMeans(n_clusters=k, random_state=1)
    kmeans.fit(data)
    #alpha_not=mu_not*alsum
    data_lol = data.values.tolist()
    data_cwise = [[] for i in range(k)]
    for i in range(n):
        data_cwise[kmeans.labels_[i]].append(data_lol[i])
    pi_not = [len(data_cwise[m])/n for m in range(k)]
    alpha_new = []
    for i in range(k):
        data_cwise_ar = np.array(data_cwise[i])
        E = data_cwise_ar.mean(axis=0)
        E2 = (data_cwise_ar ** 2).mean(axis=0)
        E3 = ((E[0] - E2[0]) / (E2[0] - E[0] ** 2)) * E
        alpha_new.append(E3.tolist())
    return pi_not, alpha_new


def gmm_init_adv(data, k):
    n = len(data)
    clf = mixture.GaussianMixture(n_components=k, covariance_type='full')
    clf.fit(data)
    #alpha_not=mu_not*alsum
    data_lol = data.values.tolist()
    data_cwise = [[] for i in range(k)]
    for i in range(n):
        data_cwise[clf.predict(data)[i]].append(data_lol[i])
    pi_not = [len(data_cwise[m])/n for m in range(k)]
    alpha_new = []
    for i in range(k):
        data_cwise_ar = np.array(data_cwise[i])
        E = data_cwise_ar.mean(axis=0)
        E2 = (data_cwise_ar ** 2).mean(axis=0)
        E3 = ((E[0] - E2[0]) / (E2[0] - E[0] ** 2)) * E
        alpha_new.append(E3.tolist())
    return pi_not, alpha_new


def dmm_loglikelihood(pi_temp, alpha_temp, data_lol):
    n = len(data_lol)
    k = len(alpha_temp)
    log_likelihood_values_temp = []
    for c in range(n):
        try:
            log_likelihood_old_temp = math.log(np.nansum(
                [pi_temp[f]*dirichlet.pdf(data_lol[c], alpha_temp[f]) for f in range(k)]))
        except:
            log_likelihood_old_temp = math.log1p(np.nansum(
                [pi_temp[f]*dirichlet.pdf(data_lol[c], alpha_temp[f]) for f in range(k)]))
        log_likelihood_values_temp.append(log_likelihood_old_temp)
    log_likelihood_old = np.sum(log_likelihood_values_temp)

    return log_likelihood_old


def dmm_responsibilities(pi_temp, alpha_temp, data_lol):
    n = len(data_lol)
    k = len(alpha_temp)
    gamma_temp = []
    for i in range(n):
        gamma_numer = []
        for j in range(k):
            temp_gamma_numer = (
                pi_temp[j]*dirichlet.pdf(data_lol[i], alpha_temp[j]))
            gamma_numer.append(temp_gamma_numer)
        gamma_row = gamma_numer / np.nansum(gamma_numer)
        gamma_temp.append(gamma_row)
    gamma_temp_ar = np.array(gamma_temp, dtype=np.float64)
    gamma_matrix = []
    for v in gamma_temp:
        gm_temp = v.tolist()
        gamma_matrix.append(gm_temp)
    return gamma_temp_ar, gamma_matrix


def dmm_pi_estimate(gamma_temp_ar):
    n = gamma_temp_ar.shape[0]
    k = gamma_temp_ar.shape[1]
    pi_new = []
    nk = []
    for g in range(k):
        nk_temp = np.nansum([gamma_temp_ar[w, g] for w in range(n)])
        pi_temp = nk_temp/n
        pi_new.append(pi_temp)
        nk.append(nk_temp)
    return pi_new


def count_to_comp(df):
    df_array=np.array(df)


    nf  = conorm.tmm_norm_factors(df)["norm.factors"]

    lj=[]
    for j in range(df_array.shape[1]):
        lj_temp=nf[j]*np.sum(df_array[:, j])
        lj.append(lj_temp)
        

    sj=[]
    for j in range(df_array.shape[1]):
        sj_temp=lj[j]/(np.sum(lj)/df_array.shape[1])
        sj.append(sj_temp)

    x_lol=[]
    for i in range(df_array.shape[0]):
        xi=[]
        for j in range(df_array.shape[1]):
            xi_temp=df_array[i,j]/sj[j]
            xi.append(xi_temp)
        xi_sum=np.sum(xi)
        xi_trans=[xi[k]/xi_sum for k in range(df_array.shape[1])]
       
            
        x_lol.append(xi_trans)
        #x_lol.append(xi)
        
    data=pd.DataFrame(x_lol)
    trans_data=pd.DataFrame(multiplicative_replacement(data))
    
    return trans_data




def dirichlet_covariance(alpha):
    p = len(alpha)
    alpha0 = np.sum(alpha)
    cov = np.zeros((p, p))

    for i in range(p):
        for j in range(p):
            if i == j:
                cov[i, j] = (alpha[i] * (alpha0 - alpha[i])) / (alpha0**2 * (alpha0 + 1))
            else:
                cov[i, j] = -(alpha[i] * alpha[j]) / (alpha0**2 * (alpha0 + 1))

    return cov



import numpy as np
import pandas as pd
import itertools
import math
from sklearn.cluster import KMeans

def fmm_loglikelihood(pi_temp,alpha_temp,data_lol,dist_comb):
    n = len(data_lol)
    k = len(alpha_temp)
    log_likelihood_values_temp = []
    for c in range(n):
        try:
            log_likelihood_old_temp = math.log(np.nansum(
                [pi_temp[f]*dist_comb[f].pdf(np.reshape(np.array(data_lol[c]),(1,len(data_lol[c]))), *alpha_temp[f]) for f in range(k)]))
        except:
            log_likelihood_old_temp = math.log1p(np.nansum(
                [pi_temp[f]*dist_comb[f].pdf(np.reshape(np.array(data_lol[c]),(1,len(data_lol[c]))), *alpha_temp[f]) for f in range(k)]))
        log_likelihood_values_temp.append(log_likelihood_old_temp)
    log_likelihood_old = np.sum(log_likelihood_values_temp)

    return log_likelihood_old


def fmm_responsibilities(pi_temp, alpha_temp, data_lol,dist_comb):
    n = len(data_lol)
    k = len(alpha_temp)
    gamma_temp = []
    for i in range(n):
        gamma_numer = []
        for j in range(k):
            temp_gamma_numer = (
                pi_temp[j]*dist_comb[j].pdf(np.reshape(np.array(data_lol[i]),(1,len(data_lol[i]))), *alpha_temp[j]))
            gamma_numer.append(temp_gamma_numer)
        gamma_row = gamma_numer / np.nansum(np.asanyarray(gamma_numer))
        gamma_temp.append(gamma_row)
    gamma_temp_ar = np.array(gamma_temp, dtype=np.float64)
    gamma_matrix = []
    for v in gamma_temp:
        gm_temp = v.tolist()
        gamma_matrix.append(gm_temp)
    return gamma_temp_ar, gamma_matrix


def fmm_pi_estimate(gamma_temp_ar):
    n = gamma_temp_ar.shape[0]
    k = gamma_temp_ar.shape[1]
    pi_new = []
    nk = []
    for g in range(k):
        nk_temp = np.nansum([gamma_temp_ar[w, g] for w in range(n)])
        pi_temp = nk_temp/n
        pi_new.append(pi_temp)
        nk.append(nk_temp)
    return pi_new


def fmm_kmeans_init(data, k,dist_comb):
    n = len(data)
    kmeans = KMeans(n_clusters=k, random_state=1)
    kmeans.fit(data)
    data_lol = data.values.tolist()
    data_cwise = [[] for i in range(k)]
    for i in range(n):
        data_cwise[kmeans.labels_[i]].append(data_lol[i])
    pi_not = [len(data_cwise[m])/n for m in range(k)]
    alpha_not=[dist_comb[j].fit(np.array(data_cwise[j])) for j in range(k)]
    return pi_not, alpha_not


def fmm_estimate_alphas(data_cwise, alpha_not, dist_comb):
    alpha_new = []
    for t in range(len(data_cwise)):
        if np.array(data_cwise[t]).ndim == 1:
            alpha_new_temp = alpha_not[t]
        else:
            alpha_new_temp = dist_comb[t].fit(np.array(np.array(data_cwise[t])))

        alpha_new.append(alpha_new_temp)
    return alpha_new

def fmm_aic(k,alpha_new,log_likelihood_new):
    return 2*((k-1)+np.sum([len(alpha_new[j]) for j in range(k)])) - 2*log_likelihood_new

def fmm_bic(k,alpha_new,log_likelihood_new,n):
    return ((k-1)+np.sum([len(alpha_new[j]) for j in range(k)]))*np.log(n) - 2*log_likelihood_new

